﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace CommunityDetection
{
	public class BeamPartition
	{
		//indices of rows belonging to this pattern
		public List<int> rows;

		//description data matrix
		BinaryMatrix data;

		//set of all rows (whole dataset, not only in partition) with classlabel 1
		HashSet<int> trueRows;
				
		//P( c=1 )
		public double p;

		public double entropy;

		public BeamNode splitNode;

		public BeamPartition(BinaryMatrix data, HashSet<int> trueRows, List<int> rows)
		{
			//save references
			this.data = data;
			this.trueRows = trueRows;
			this.rows = rows;
			this.rows = rows;

			splitNode = new BeamNode();

			//calculate the fraction with classlabel 1 in this paritition
			p = 0;
			foreach (int r in rows)
			{
				if (trueRows.Contains(r))
				{
					splitNode.nt++;
					p++;
				}
				else
				{
					splitNode.nf++;
				}
			}
			p /= rows.Count;

			//calculate entropy of this partition
			entropy = p == 0 || p == 1 ? 0 : (-p * log(p) - (1 - p) * log(1 - p));

			splitNode.classlabel = p > 0.5;
		}

		/** 
		 * Calculate the conditional entropy based on a given theta and q
		 * @param theta P( x=1) (support dataset)
		 * @param q P( c=1 | x=1 ) (supp
		 */
		public double conditionalEntropy(double theta, double q)
		{
			//formula for conditional entropy as given in "Discriminative Frequent Pattern Analasys for Effective Classification"
			//treat x log x with x = 0 as 0
			
			double hcx1,hcx2,hcx3;

			if (theta == 0 || q == 0 || q == 1)
				hcx1 = 0;
			else
				hcx1 = -theta * q * log(q) - theta * (1 - q) * log(1 - q);
			
			if (theta * q == p || theta == 1)
				hcx2 = 0;
			else
				hcx2 = (theta * q - p) * log((p - theta * q) / (1 - theta));

			hcx3 = (theta * (1 - q) - (1 - p)) * log(((1 - p) - theta * (1 - q)) / (1 - theta));
			
			//divide by 0, or log 0, always means there is some 0 * log(0) or * 0 * (x/0)
			if (double.IsNaN(hcx3))
				hcx3 = 0;

			return hcx1 + hcx2 + hcx3;
		}

		public BeamPattern findBestPattern(int width, int maxdepth, HashSet<int> attributes)
		{
			//init
			int depth = 0;
			List<BeamPattern> beam = new List<BeamPattern>();
			BeamPattern best = new BeamPattern();
			foreach (int r in rows)
				best.rows = rows;
			beam.Add(best);

			while (depth < maxdepth)
			{
				HashSet<BeamPattern> newBeam = new HashSet<BeamPattern>();

				foreach (BeamPattern pattern in beam)
				{
					foreach (int a in attributes)
					{
						//check if att is part of pattern
						if (pattern.pattern.Contains(a))
							continue;				

						//create new pattern
						BeamPattern newpat = new BeamPattern();
						newpat.AddAttribute(a);
						foreach (int pa in pattern.pattern)
							newpat.AddAttribute(pa);
						if (newBeam.Contains(newpat))
							continue;

						double cp = 0; //contains pattern, class 1
						double cn = 0; //contains pattern, class 0
						double n = 0; //does not contain pattern;

						foreach (int r in pattern.rows) //only check pattern rows, other rows cant contain new pattern because it is superset of "pattern"
						{
							if (data.getRow(r).contains(newpat.pattern))
							{
								newpat.AddRow(r);
								if (trueRows.Contains(r))
									cp++;
								else
									cn++;
							}
							else
							{
								n++;
							}
						}

						//formula page  4 "Discriminative frequent pattern analysis for effective classification"
						double theta = (cn + cp) / rows.Count; // support pattern

						//check also if there are rows (c=1 | newpat)
						double q = cp != 0 ? cp / (cn + cp) : 0; // P( c=1 | newpat)

						//calculate information gain
						newpat.gain = entropy - conditionalEntropy(theta, q);

						//add to new beam
						newBeam.Add(newpat);
					}
				}

				//sort new beam by information gain
				beam = new List<BeamPattern>(newBeam);
				beam.Sort();

                // if there's anything in the beam
                if (beam.Count > 0)
                {
                    //best one improved?
                    if (best.gain < beam[0].gain)
                    {
                        best = beam[0];
                    }
                    beam.RemoveRange(Math.Min(beam.Count, width) - 1, Math.Max(beam.Count - width, 0));
                }
				// = new List<BeamPattern>(newBeam.Take(width));
				//Console.WriteLine("depth: " + depth);
				//for (int i = 0; i < beam.Count; i++)
				//	Console.WriteLine(i + ": " + beam[i]);
				depth++;
			}
			return best;
		}

		/**
		* Split this parition on a certain pattern
		*/
		public List<BeamPartition> split(BeamPattern pattern)
		{
			//split according to new pattern
			List<int> rowsT = new List<int>();
			List<int> rowsF = new List<int>();
			foreach (int r in rows)
			{
				BitArray row = data.getRow(r);
				bool contains = true;
				foreach (int i in pattern.pattern)
				{
					if (!row.get(i))
					{
						contains = false;
						break;
					}
				}

				if (contains)
					rowsT.Add(r);
				else
					rowsF.Add(r);
			}

			List<BeamPartition> splitresult = new List<BeamPartition>();
			if (rowsT.Count == 0 || rowsF.Count == 0)
			{
				splitresult.Add(this);
				return splitresult;
			}

			//create two partitions
			BeamPartition t = new BeamPartition(data, trueRows, rowsT);
			BeamPartition f = new BeamPartition(data, trueRows, rowsF);

			//set pattern for splitnode
			splitNode.splitPattern = pattern;

			//create left subtree
			splitNode.left = f.splitNode;
			f.splitNode.parent = splitNode;

			//create right subtree
			splitNode.right = t.splitNode;
			t.splitNode.parent = splitNode;

			splitresult.Add(t);
			splitresult.Add(f);
			return splitresult;
		}

		public List<BeamPartition> split(List<BeamPattern> patterns)
		{
			Dictionary<BeamNode, List<int>> map = new Dictionary<BeamNode, List<int>>();
			foreach (int r in rows)
			{
				BitArray row = data.getRow(r);
				BeamNode current = splitNode;
				for(int i=0; i<patterns.Count; i++)
				{
					BeamPattern pattern = patterns[i];
					if (current.splitPattern == null)
					{
						current.splitPattern = pattern;
						current.left = new BeamNode();
						current.right = new BeamNode();
						current.left.parent = current;
						current.right.parent = current;
					}

					if (row.contains(pattern.pattern)) //right
						current = current.right;
					else
						current = current.left;
	
					if (trueRows.Contains(r))
						current.nt++;
					else
						current.nf++;

					current.classlabel = current.nt > current.nf;


				}
				if (!map.ContainsKey(current))
					map[current] = new List<int>();
				map[current].Add(r);
			}

			List<BeamPartition> partitions = new List<BeamPartition>();
			foreach (KeyValuePair<BeamNode, List<int>> kv in map)
			{
				BeamPartition p = new BeamPartition(data, trueRows, kv.Value);
				p.splitNode = kv.Key;
				partitions.Add(p);
			}
			return partitions;
		}

		public double log(double d)
		{
			return Math.Log(d, 2);
		}

		public override string ToString()
		{
			return "size: "+this.rows.Count+", P(C=1):"+this.p+", sp:"+splitNode;
		}
	}
}
