﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace CommunityDetection
{
	class DecisionNode
	{
		public int numT, numF;
		public int[] columnSum;
		public double imp;
		BinaryMatrix data;

		public DecisionNode leftNode, rightNode, parent;
		DecisionTree tree;

		public bool isLeaf;
		public int splitAttribute;
		public double alpha;
		public int numLeaves;

		public DecisionNode(DecisionTree tree, DecisionNode parent, BinaryMatrix data, int[] columnSum)
		{
			this.data = data;
			this.columnSum = columnSum;
			splitAttribute = -1;
			isLeaf = true;
			this.tree = tree;
			this.parent = parent;
		}



		public DecisionNode(DecisionTree tree,DecisionNode node, DecisionNode parent)
		{
			//new
			this.tree = tree;
			this.parent = parent;

			//copy
			this.numT = node.numT;
			this.numF = node.numF;
			this.columnSum = node.columnSum;
			this.imp = node.imp;
			this.data = node.data;
			this.isLeaf = node.isLeaf;
			this.splitAttribute = node.splitAttribute;
			this.alpha = node.alpha;
			this.numLeaves =node.numLeaves;

			//recursion
			if (!isLeaf)
			{
				this.leftNode = new DecisionNode(tree, node.leftNode, this);
				this.rightNode = new DecisionNode(tree, node.rightNode, this);
			}
		}

		public int getSubtreeSize()
		{
			if (isLeaf)
				return 0;
			else
				return leftNode.getSubtreeSize() + rightNode.getSubtreeSize() + 1;
		}

		public double calcNodeCost()
		{
			double cases = tree.getRoot().numT + tree.getRoot().numF;
			return resubsitutionErrorNode() * ((double)(numT + numF)) / cases;
		}

		public double resubsitutionErrorNode()
		{
			return (1.0 * Math.Min(numF, numT)) / (numF + numT);
		}

		public double fraction()
		{
			return ((double)(numF + numT)) / (tree.getRoot().numT + tree.getRoot().numF);
		}
		public  List<DecisionNode> calculateAlpha()
		{
			if (isLeaf)
			{
				List<DecisionNode> list = new List<DecisionNode>();
				list.Add(this);
				alpha = -1;
				numLeaves = 1;
				return list;
			}
			else
			{
				double sum = 0;
				List<DecisionNode> list = leftNode.calculateAlpha();
				foreach (DecisionNode n in list)
					sum += n.resubsitutionErrorNode()*n.fraction();
				foreach (DecisionNode n in rightNode.calculateAlpha())
				{
					sum += n.resubsitutionErrorNode() * n.fraction();
					list.Add(n);
				}
				numLeaves = list.Count;
				alpha = (resubsitutionErrorNode() * fraction() - sum) / (list.Count - 1);
				return list;
			}
		}

		public void getLeavesRecursive(List<DecisionNode> list)
		{
			if (isLeaf)
			{
				list.Add(this);
			}
			else
			{
				leftNode.getLeavesRecursive(list);
				rightNode.getLeavesRecursive(list);
			}
		}

		public bool classify(BitArray values)
		{
			if (isLeaf)
				return numT > numF;
			if (!values.get(splitAttribute))
				return leftNode.classify(values);
			return rightNode.classify(values);
		}

		public double splitGain(int a, int numRows, HashSet<int> trueRows)
		{
			int atct = 0;
			foreach (int v in trueRows)
			{
				if (data.get(v, a))
					atct++;
			}
			int afct = trueRows.Count - atct;
			int atcf = columnSum[a] - atct;
			int afcf = numRows - columnSum[a] - afct;

			double fracL = ((double)(afct + afcf)) / numRows;
			double fracR = ((double)(atct + atcf)) / numRows;

			ImpurityMeasure measure = tree.Measure;
			double dImp = imp - (fracL * measure.impurity(afct, afcf) + fracR * measure.impurity(atct, atcf));
			return dImp;
		}

		public void split(List<int> attributes, HashSet<int> rows, HashSet<int> trueRows, double imp)
		{
			this.imp = imp;
			this.numT = trueRows.Count;
			this.numF = rows.Count - trueRows.Count;
			
			//best split
			int bestSplitAttribute = -1;
			double bestSplitGain = double.MinValue;
			foreach (int a in attributes)
			{
				double gain = splitGain(a, rows.Count, trueRows);
				if (gain > bestSplitGain)
				{
					bestSplitGain = gain;
					bestSplitAttribute = a;
				}
			}

			if (bestSplitGain > 0) //split is useful
			{
				
				splitAttribute = bestSplitAttribute;
				attributes.Remove(bestSplitAttribute);
				isLeaf = false;
				
				//attributes
				List<int> attributesR = new List<int>(attributes);
				attributes = null; //memspace
				List<int> attributesL = new List<int>(attributesR);
				
				//rows
				HashSet<int> rowsR = new HashSet<int>();
				HashSet<int> rowsL = new HashSet<int>();
				foreach (int v in rows)
				{
					if (data.get(v, splitAttribute))
						rowsR.Add(v);
					else
						rowsL.Add(v);
				}
				rows = null; //memspace

				//true rows
				HashSet<int> trueRowsR = new HashSet<int>(rowsR.Intersect(trueRows));
				HashSet<int> trueRowsL = new HashSet<int>(rowsL.Intersect(trueRows));
				trueRows = null; // memspace
				
				//columnsum
				int[] columnSumL = new int[columnSum.Length];
				int[] columnSumR = new int[columnSum.Length];

				if (rowsR.Count > rowsL.Count)
				{
					foreach (int i in rowsL)
					{
						foreach (int a in attributesL)
						{
							if (data.get(i, a))
								columnSumL[a]++;
						}
					}
					foreach (int a in attributesL)
					{
						columnSumR[a] = columnSum[a] - columnSumL[a];
					}
				}
				else
				{
					foreach (int i in rowsR)
					{
						foreach (int a in attributesR)
						{
							if (data.get(i, a))
								columnSumR[a]++;
						}
					}
					foreach (int a in attributesL)
					{
						columnSumL[a] = columnSum[a] - columnSumR[a];
					}
				}

				leftNode = new DecisionNode(tree, this, data, columnSumL);
				rightNode = new DecisionNode(tree, this, data, columnSumR);
				leftNode.split(attributesL, rowsL, trueRowsL);
				rightNode.split(attributesR, rowsR, trueRowsR);
			}
			else //leaf
			{
				isLeaf = true;
			}
		}
		/**
		* For root node calulate imp
		*/
		public void split(List<int> attributes, HashSet<int> rows, HashSet<int> trueRows)
		{
			split(attributes, rows, trueRows, tree.Measure.impurity(trueRows.Count, rows.Count - trueRows.Count));
		}

		public bool ClassLabel
		{
			get
			{
				return numT > numF;
			}
		}
		public override string ToString()
		{
			return (parent==null?"Root - ":"")+(isLeaf?ClassLabel+" leaf":"Internal")+" - Impurity: " + imp + ", #T:" + numT + ", #F:" + numF;
		}

		public void print()
		{
			print(Console.Out);
		}

		public void print(TextWriter writer)
		{
			print(0, "", writer);
		}
		
		private void print(int d, string v, TextWriter writer)
		{
			string sum = "";
			for (int i = 0; columnSum != null && i < columnSum.Length; i++)
			{
				sum += columnSum[i];
			}
			writer.WriteLine(new String(' ', d * 2) + (v.Equals("") ? "" : v + " - ") + (isLeaf?"":"split: " + splitAttribute + ", leaves:"+numLeaves+", impurity: ") + imp + ", instances(T+F):"+numT+"+"+numF+"=" + (numF+numT) + ", type:" + (isLeaf ? " leaf (" + (this.numF>numT?"False":"True") + ")" : " internal"));
			if (!isLeaf)
			{
				leftNode.print(d + 1, "F", writer);
				rightNode.print(d + 1, "T", writer);
			}
		}
	}
}
